{% extends "basic_template.html" %}
{% block title %}
{{ super() }}
{% endblock %}
{% block head %}
{{ super() }}
{% endblock %}
{% block content %}

<h1>404</h1>

Whoops! Back to the <a href="/">main page.</a>

{% endblock %}
