{% extends "basic_template.html" %}
{% block title %}
{{ super() }}
{% endblock %}
{% block head %}
{{ super() }}
{% endblock %}
{% block content %}

<h1>Source</h1>


<section class="w3-container w3-margin">
  <article>
    <aside class="w3-container w3-left">
      <center>
        <a href="https://github.com/knappa/steadycellphenotype">
          <img src="{{ url_for('static', filename='GitHub-Mark-64px.png') }}"
               alt="Github"
               width="64px"
               height="64px" >
        </a>
      </center>
    </aside>
    <p>
      The source code for this web site and associated utilities are licensed under the
      <a rel="license" href="https://github.com/knappa/steadycellphenotype/blob/master/LICENSE">MIT
      License</a> and is available
      on <a href="https://github.com/knappa/steadycellphenotype">Github</a>.
    </p>
    <p>
      The authors ask that anyone publishing works using, or derived from, this simulator cite our
      paper, currently in preprint form at:
      <cite>
        <a href="https://doi.org/10.1101/2021.08.10.454961">
          doi:10.1101/2021.08.10.454961
        </a>
      </cite>
    </p>
  </article>
</section>

{% endblock %}
